class ArtifactNotFound(Exception):
    ...


class CorruptedArtifact(Exception):
    ...


class CorruptedMetadata(Exception):
    ...


class CorruptedONNXModel(Exception):
    ...


class NoSuchModel(Exception):
    ...


class SelectModelRequired(Exception):
    ...



import copy
from collections import OrderedDict
from typing import List, Optional, Sequence, Tuple, Union

import numpy as np
import torch
import torch.nn as nn

from .display import DisplayMixin


def is_seq_int(tp) -> bool:
    """Return True if the input is a sequence of integers."""
    return tp and isinstance(tp, Sequence) and all(isinstance(p, int) for p in tp)


class Tailor(DisplayMixin):
    def __init__(
        self,
        model: nn.Module,
        input_shape: Optional[Tuple[int, ...]] = None,
        input_dtype: str = 'float32',
    ):
        """Tailor converts a general DNN model into an embedding model.

        :param model: a general DNN model
        :param input_shape: a sequence of integers defining the shape of the input
            tensor. Note, batch size is *not* part of ``input_shape``. It is required
            for :py:class:`Tailor`.
        :param input_dtype: the data type of the input tensor.
        """
        self._model = model

        # multiple inputs to the network
        if isinstance(input_shape, tuple):
            input_shape = [input_shape]

        self._input_shape = input_shape
        self._input_dtype = input_dtype

    def to_embedding_model(
        self,
        layer_name: Optional[str] = None,
        freeze: Union[bool, List[str]] = False,
        projection_head: Optional[nn.Module] = None,
    ) -> nn.Module:
        """Convert a general model from :py:attr:`.model` to an embedding model.

        :param layer_name: the name of the layer that is used for output embeddings.
            All layers *after* that layer will be removed. When set to ``None``, then
            the last layer listed in :py:attr:`.embedding_layers` will be used. To see
            all available names you can check ``name`` field of
            :py:attr:`.embedding_layers`.
        :param freeze: if set as True, will freeze all layers before
            :py:`attr`:`layer_name`. If set as list of str, will freeze layers by names.
        :param projection_head: Attach a module at the end of model, this module should
            be always trainable.
        :return: Converted embedding model.
        """

        model = copy.deepcopy(self._model)
        _all_embed_layers = {layer['name']: layer for layer in self.embedding_layers}
        if layer_name:
            try:
                _embed_layer = _all_embed_layers[layer_name]
            except KeyError as e:
                raise KeyError(
                    f'`embedding_layer_name` must be one of {_all_embed_layers.keys()}'
                    f', given {layer_name}'
                ) from e
        else:
            # when not given, using the last layer
            _embed_layer = self.embedding_layers[-1]

        if isinstance(freeze, list):
            # freeze specific layers defined in `freeze_layers`
            for layer_name, param in zip(_all_embed_layers, model.parameters()):
                if layer_name in freeze:
                    param.requires_grad = False
        elif isinstance(freeze, bool) and freeze is True:
            # freeze all layers, not including bottleneck module
            for param in model.parameters():
                param.requires_grad = False

        _embed_layer_output_shape = 0
        _relative_idx_to_embedding_layer = None
        for name, module in model.named_modules():
            if name == _embed_layer['module_name']:
                _relative_idx_to_embedding_layer = 0
                _embed_layer_output_shape = _embed_layer['output_shape']
            if (
                _relative_idx_to_embedding_layer
                and _relative_idx_to_embedding_layer >= 1
            ):
                replaced_layer = nn.Identity(_embed_layer_output_shape)
                if '.' in name:
                    # Note: in torchvision, nested layer names are named with '.'
                    # e.g. classifier.0
                    nested_module, layer = name.split('.')
                    setattr(getattr(model, nested_module), layer, replaced_layer)
                else:
                    setattr(model, name, replaced_layer)

            if _relative_idx_to_embedding_layer is not None:
                _relative_idx_to_embedding_layer += 1

        if projection_head:
            embed_model_with_projection_head = nn.Sequential()
            embed_model_with_projection_head.add_module('embed_model', model)
            embed_model_with_projection_head.add_module(
                'projection_head', projection_head
            )
            return embed_model_with_projection_head

        return model

    @property
    def embedding_layers(self) -> List[dict]:
        """Get all dense layers that can be used as embedding layer from the
        :py:attr:`.model`.

        :return: layers info list of dicts.
        """
        _layers = self.summary()
        return [_l for _l in _layers if _l['is_embedding_layer']]

    def summary(self, skip_identity_layer: bool = False) -> List[dict]:
        """Interpret the DNN model and produce model information.

        :param skip_identity_layer: If skip identity layer.
        :return: The model information stored as list of dicts.
        """
        if not self._input_shape:
            raise ValueError(
                f'{self.__class__} requires a valid `input_shape`, but receiving '
                f'{self._input_shape}'
            )

        user_model = copy.deepcopy(self._model)
        dtypes = [getattr(torch, self._input_dtype)] * len(self._input_shape)
        depth = len(list(user_model.modules()))
        for name, module in user_model.named_modules():
            module.name = name

        def _get_shape(output):
            output_shape = None
            if output is not None:
                if isinstance(output, (list, tuple)):
                    output_shape = [
                        _get_shape(o) for o in output if isinstance(o, torch.Tensor)
                    ]
                    if len(output_shape) == 1:
                        output_shape = output_shape[0]
                elif isinstance(output, torch.Tensor):
                    output_shape = list(output.shape)

            return output_shape

        def register_hook(module):
            def hook(module, input, output):
                input_shape = _get_shape(input)
                output_shape = _get_shape(output)

                if input_shape and output_shape:
                    class_name = str(module.__class__).split('.')[-1].split("'")[0]

                    module_idx = len(summary)

                    m_key = f'{class_name.lower()}_{module_idx + 1}'
                    summary[m_key] = OrderedDict()
                    summary[m_key]['cls_name'] = module.__class__.__name__
                    summary[m_key]['name'] = m_key
                    summary[m_key]['output_shape'] = output_shape
                    summary[m_key]['input_shape'] = input_shape
                    summary[m_key]['module_name'] = module.name

                    params = 0
                    summary[m_key]['trainable'] = False
                    if hasattr(module, 'weight') and hasattr(module.weight, 'size'):
                        params += np.prod(list(module.weight.size()))
                        summary[m_key]['trainable'] = module.weight.requires_grad
                    if hasattr(module, 'bias') and hasattr(module.bias, 'size'):
                        params += np.prod(list(module.bias.size()))
                    if hasattr(module, 'all_weights'):
                        params += sum(
                            np.prod(ww.size()) for w in module.all_weights for ww in w
                        )

                    summary[m_key]['nb_params'] = params

            if (
                not isinstance(module, nn.Sequential)
                and not isinstance(module, nn.ModuleList)
                and (module != user_model or depth < 1)
            ):
                hooks.append(module.register_forward_hook(hook))

        device = next(user_model.parameters()).device
        x = [
            torch.randint(0, 2, (2, *in_shape), device=device, dtype=dt)
            for in_shape, dt in zip(self._input_shape, dtypes)
        ]

        # create properties
        summary = OrderedDict()
        hooks = []

        # register hook
        user_model.apply(register_hook)

        # make a forward pass
        user_model(*x)

        # remove these hooks
        for h in hooks:
            h.remove()

        results = []
        for idx, layer in enumerate(summary):
            output_shape = summary[layer]['output_shape']
            input_shape = summary[layer]['input_shape']
            is_embedding_layer = not (
                not output_shape
                or not is_seq_int(output_shape)
                or summary[layer]['cls_name'] == self._model.__class__.__name__
            )

            if (
                skip_identity_layer
                and output_shape == input_shape
                and not summary[layer]['nb_params']
            ):
                # not an effective layer, often a wrapper/identity layer
                continue

            results.append(
                {
                    **summary[layer],
                    'output_features': output_shape[-1],
                    'output_shape_display': output_shape[1:],
                    'layer_idx': idx,
                    'is_embedding_layer': is_embedding_layer,
                }
            )

        return results



import io
import logging
import multiprocessing as mp
import os
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union

import numpy as np
import onnxruntime as ort
import torch
from docarray import DocumentArray
from rich.logging import RichHandler

from .excepts import CorruptedMetadata, CorruptedONNXModel
from .load import load_artifact

from . import collate, preprocess, utils

logging.basicConfig(
    level='DEBUG',
    format='%(message)s',
    datefmt='[%X]',
    handlers=[RichHandler()],
)
logger = logging.getLogger(__name__)


class ONNXRuntimeInferenceEngine:
    _artifact: str
    _token: Optional[str]
    _batch_size: int
    _select_model: str
    _device: str
    _device_id: int
    _omp_num_threads: int
    _intra_op_num_threads: int
    _inter_op_num_threads: int
    _logging_level: str

    _model_name: str
    _metadata: Dict[str, Any]
    _onnx_dump: Union[None, str, io.BytesIO]
    _input_names: List[str]
    _input_shapes: List[Tuple[Union[str, int]]]
    _output_name: str
    _output_shape: Tuple[Union[int, str]]
    _multi_modal: bool
    _datahandler: utils.PreprocessCollateWrapper

    # noinspection PyUnresolvedReferences
    _session_options: ort.SessionOptions
    _providers: List[str]
    _session: ort.InferenceSession

    def __init__(
        self,
        artifact: str,
        token: Optional[str] = None,
        batch_size: int = 32,
        select_model: Optional[str] = None,
        device: str = 'cpu',
        device_id: int = 0,
        omp_num_threads: int = mp.cpu_count(),
        intra_op_num_threads: int = 0,
        inter_op_num_threads: int = 0,
        logging_level: str = 'DEBUG',
        **kwargs,
    ) -> None:
        """
        The finetuner executor can load any model trained with Finetuner and use it to
        encode Documents.

        :param artifact: Specify a finetuner run artifact. Can be a path to a local
            directory, a path to a local zip file or a Hubble artifact ID. Individual
            model artifacts (model subfolders inside the run artifacts) can also be
            specified using this argument.
        :param token: A Jina authentication token required for pulling artifacts from
            Hubble. If not provided, the Hubble client will try to find one either in a
            local cache folder or in the environment.
        :param batch_size: Incoming documents are fed to the graph in batches, both to
            speed-up inference and avoid memory errors. This argument controls the
            number of documents that will be put in each batch.
        :param select_model: Finetuner run artifacts might contain multiple models. In
            such cases you can select which model to deploy using this argument.
        :param device: The device to use for inference. Either `cpu` or `cuda`.
        :param device_id: Specify which CPU or GPU to use for inference.
        :param omp_num_threads: The number of threads set by OpenMP. Check out
            https://www.openmp.org/spec-html/5.0/openmpse50.html for more information.
            By default it is set to the number of available CPUs.
        :param intra_op_num_threads: The execution of an individual operation (for some
            operation types) can be parallelized on a pool of `intra_op_num_threads`. 0
            means the system picks an appropriate number.
            See https://onnxruntime.ai/docs/api/python/api_summary.html#onnxruntime.SessionOptions.intra_op_num_threads
        :param inter_op_num_threads: Nodes that perform blocking operations are enqueued
            on a pool of `inter_op_num_threads` available in each process. 0 means
            the system picks an appropriate number.
            See https://onnxruntime.ai/docs/api/python/api_summary.html#onnxruntime.SessionOptions.inter_op_num_threads
        :param logging_level: The executor logging level. See
            https://docs.python.org/3/library/logging.html#logging-levels for available
            options.
        """

        super().__init__(**kwargs)

        self._artifact = artifact
        self._token = token
        self._batch_size = batch_size
        self._select_model = select_model
        self._device = device.lower().strip()
        self._device_id = device_id
        self._omp_num_threads = omp_num_threads
        self._intra_op_num_threads = intra_op_num_threads
        self._inter_op_num_threads = inter_op_num_threads
        self._logging_level = logging_level

        # noinspection PyUnresolvedReferences
        assert self._logging_level in logging._nameToLevel
        logger.setLevel(self._logging_level)

        logger.info('Starting the executor ...')
        logger.debug(f'Artifact set to [{self._artifact}]')
        logger.debug(f'Device set to [{self._device}]')
        logger.debug(f'Device ID set to [{self._device_id}]')
        logger.debug(f'Batch size set to [{self._batch_size}]')
        logger.debug(f'OpenMP num threads set to [{self._omp_num_threads}]')
        logger.debug(
            f'Intra operation num threads set to [{self._intra_op_num_threads}]'
        )
        logger.debug(
            f'Inter operation num threads set to [{self._inter_op_num_threads}]'
        )
        logger.debug(f'Logging level set to [{self._logging_level}]')

        self._load_artifact()
        self._check_metadata()
        self._build_data_pipeline()
        self._configure_onnx_runtime()
        self._build_onnx_graph()

    """ Start-up methods """

    def _load_artifact(self):
        """Load the artifact either from a directory, a zip file or directly from
        Hubble."""

        logger.info('Loading artifact ...')
        self._model_name, self._metadata, self._onnx_dump = load_artifact(
            self._artifact, self._select_model, self._token
        )

    def _check_metadata(self):
        """Check the model metadata."""

        logger.info('Checking the model metadata ...')

        try:
            self._input_names = self._metadata['input_names']
            self._input_shapes = self._metadata['input_shapes']
            self._output_name = self._metadata['output_name']
            self._output_shape = self._metadata['output_shape']
        except KeyError as exc:
            raise CorruptedMetadata(f'Key is missing, details: {str(exc)}')

        if len(self._input_names) != len(self._input_shapes):
            raise CorruptedMetadata(
                'Length mismatch between keys \'input_names\' and \'input_shapes\''
            )

        for name, shape in zip(self._input_names, self._input_shapes):
            logger.debug(f'Input {name}: {shape}')

        logger.debug(f'Output: {self._output_shape}')

        _input_dynamic_axes = {
            dim
            for _input_shape in self._input_shapes
            for dim in _input_shape
            if isinstance(dim, str)
        }
        _output_dynamic_axes = {
            dim for dim in self._output_shape if isinstance(dim, str)
        }
        _set_difference = _output_dynamic_axes - _input_dynamic_axes
        if len(_set_difference) > 0:
            raise CorruptedMetadata(
                f'Output dynamic axes {_set_difference} not found in the inputs. '
                'Dynamic shape inference will fail'
            )

    def _build_data_pipeline(self):
        """Configure the preprocess and the collate functions from the model
        metadata."""

        logger.info('Building the data pipeline ...')

        preprocess_fns = {}
        collate_fns = {}

        try:

            modalities = list(self._metadata['preprocess_types'].keys())
            logger.debug(f'Modalities: {modalities}')

            for modality in modalities:

                preprocess_type = self._metadata['preprocess_types'][modality]
                preprocess_options = self._metadata['preprocess_options'][modality]
                collate_type = self._metadata['collate_types'][modality]
                collate_options = self._metadata['collate_options'][modality]

                preprocess_fn = getattr(preprocess, preprocess_type)(
                    **preprocess_options
                )

                collate_fn = getattr(collate, collate_type)(**collate_options)

                logger.debug(
                    f'{modality} - {preprocess_type}({preprocess_options}) '
                    f'{preprocess_type}({preprocess_options})'
                )

                preprocess_fns[modality] = preprocess_fn
                collate_fns[modality] = collate_fn

        except Exception as exc:
            raise CorruptedMetadata(
                f'Preprocess and collate are misconfigured, details: {str(exc)}'
            )

        self._multi_modal = len(modalities) > 1
        self._datahandler = utils.PreprocessCollateWrapper(
            preprocess_fns, collate_fns, self._multi_modal
        )

    def _configure_onnx_runtime(self):
        """
        Configure the ONNX Runtime options.
        """
        logger.info('Configuring the ONNX runtime ...')

        if self._device not in ['cpu', 'cuda']:
            raise ValueError(f'Unknown device type \'{self._device}\'')

        # noinspection PyUnresolvedReferences
        self._session_options = ort.SessionOptions()
        self._providers = []

        if self._device == 'cpu':

            os.environ['OMP_NUM_THREADS'] = str(self._omp_num_threads)
            self._providers = ['CPUExecutionProvider']

            # noinspection PyUnresolvedReferences
            self._session_options.execution_mode = (
                ort.ExecutionMode.ORT_SEQUENTIAL
                if self._intra_op_num_threads <= 1
                else ort.ExecutionMode.ORT_PARALLEL
            )
            self._session_options.intra_op_num_threads = self._intra_op_num_threads

            if self._inter_op_num_threads > 1:
                self._session_options.inter_op_num_threads = self._inter_op_num_threads

        else:
            self._providers = ['CUDAExecutionProvider']

    def _build_onnx_graph(self):
        """Load the ONNX model."""
        logger.info('Building the ONNX graph ...')
        try:

            if isinstance(self._onnx_dump, str):
                f = self._onnx_dump
            else:
                self._onnx_dump.seek(0)
                f = self._onnx_dump.read()
                self._onnx_dump = None

            self._session = ort.InferenceSession(
                f, self._session_options, providers=self._providers
            )
        except Exception as exc:
            raise CorruptedONNXModel(
                f'ONNX model appears to be corrupted, details: {str(exc)}'
            )

        if len(self._session.get_outputs()) != 1:
            raise CorruptedMetadata(
                'Multi-output models are not supported, '
                f'ONNX graph outputs: {self._session.get_outputs()}'
            )

        onnx_input_names = [node.name for node in self._session.get_inputs()]
        if set(onnx_input_names) != set(self._input_names):
            raise CorruptedMetadata(
                f'Invalid ONNX graph inputs {onnx_input_names}, '
                f'expected {self._input_names}'
            )

    """ Runtime methods """

    @property
    def _numpy_to_torch_dtype(self):
        # https://github.com/pytorch/pytorch/blob/ac79c874cefee2f8bc1605eed9a924d80c0b3542/torch/testing/_internal/common_utils.py#L349
        return {
            bool: torch.bool,
            np.uint8: torch.uint8,
            np.int8: torch.int8,
            np.int16: torch.int16,
            np.int32: torch.int32,
            np.int64: torch.int64,
            np.float16: torch.float16,
            np.float32: torch.float32,
            np.float64: torch.float64,
            np.complex64: torch.complex64,
            np.complex128: torch.complex128,
        }

    @property
    def _torch_to_numpy_dtype(self):
        return {v: k for k, v in self._numpy_to_torch_dtype.items()}

    def _run_data_pipeline(
        self,
        docs: DocumentArray,
    ) -> Dict[str, Union[torch.Tensor, Mapping]]:
        """Run documents through the data pipeline."""
        return self._datahandler.collate_contents(
            [self._datahandler.preprocess(doc) for doc in docs]
        )

    def _flatten_inputs(
        self,
        inputs: Dict[str, Union[torch.Tensor, Mapping]],
    ) -> Dict[str, torch.Tensor]:
        """Run documents through the data pipeline."""
        tensors = {}
        for _input_name, _input in inputs.items():
            if isinstance(_input, torch.Tensor):
                tensors[_input_name] = _input
            elif isinstance(_input, Mapping):
                for _sub_input_name, _sub_input in _input.items():
                    if not isinstance(_sub_input, torch.Tensor):
                        raise RuntimeError(
                            f'Got unexpected type {type(_sub_input)} for input '
                            f'name {_sub_input_name}'
                        )
                    tensors[_sub_input_name] = _sub_input
            else:
                raise RuntimeError(
                    f'Got unexpected type {type(_input)} for input name '
                    f'{_input_name}'
                )
        return tensors

    def _check_input_names(self, inputs: Dict[str, torch.Tensor]) -> None:
        """Run documents through the data pipeline."""
        _set_given_inputs = set(inputs.keys())
        _set_expected_inputs = set(self._input_names)
        if _set_given_inputs.intersection(_set_expected_inputs) != _set_expected_inputs:
            raise RuntimeError(
                f'Expected input names {self._input_names}, got {list(inputs.keys())}'
            )

    def _move_to_device(
        self, inputs: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """Moves tensoors to device."""
        tensors = {}
        for _input_name, _input_tensor in inputs.items():
            if _input_tensor.dtype in [torch.int64, torch.long]:
                # int32 mandatory as input of bindings, int64 not supported
                tensors[_input_name] = _input_tensor.int().to(self._device)
            else:
                tensors[_input_name] = _input_tensor.to(self._device)
        return tensors

    def _infer_output_shape(self, inputs: Dict[str, torch.Tensor]) -> Tuple[int, ...]:
        """Infer the output shape by inspecting the input tensors. Every dynamic axis
        in the output shape is replaced by the value of the same axis in the input.
        """
        _dynamic_axes: Dict[str, int] = {}
        for name, shape in zip(self._input_names, self._input_shapes):
            tensor = inputs[name]
            for _given_dim, _expected_dim in zip(tensor.shape, shape):
                if isinstance(_expected_dim, str):
                    _dynamic_axes[_expected_dim] = _given_dim
                else:
                    if _given_dim != _expected_dim:
                        raise RuntimeError(
                            f'Shape mismatch, expected {shape} got {tensor.shape}'
                        )
        return tuple(
            [
                _dynamic_axes[dim] if isinstance(dim, str) else dim
                for dim in self._output_shape
            ]
        )

    def _run_graph(
        self,
        inputs: Dict[str, torch.Tensor],
        output_shape: Tuple[int, ...],
    ) -> torch.Tensor:
        """Performs inference on ONNX Runtime in an optimized way. In particular,
        avoids some tensor copies from GPU to host by using Torch tensors
        directly."""

        binding: ort.IOBinding = self._session.io_binding()

        _inputs = {}
        for _input_name in self._input_names:
            tensor = inputs[_input_name].contiguous()
            binding.bind_input(
                name=_input_name,
                device_type=self._device,
                device_id=self._device_id,
                element_type=self._torch_to_numpy_dtype[tensor.dtype],
                shape=tuple(tensor.shape),
                buffer_ptr=tensor.data_ptr(),
            )
            _inputs[_input_name] = tensor

        output = torch.empty(
            output_shape, dtype=torch.float32, device=self._device
        ).contiguous()

        binding.bind_output(
            name=self._output_name,
            device_type=self._device,
            device_id=self._device_id,
            element_type=np.float32,  # hard coded output type
            shape=output_shape,
            buffer_ptr=output.data_ptr(),
        )

        self._session.run_with_iobinding(binding)

        return output



import abc
from typing import TYPE_CHECKING

import numpy as np
from docarray import Document

if TYPE_CHECKING:
    from docarray.typing import DocumentContentType


_CHANNEL_LAST = [-1, 2]


class BasePreprocess(abc.ABC):
    @abc.abstractmethod
    def __call__(self, doc: Document) -> 'DocumentContentType':
        ...


class DefaultPreprocess(BasePreprocess):
    """
    Default built-in preprocess class to unpack the content from a ``Document`` object.
    """

    def __call__(self, doc: Document) -> 'DocumentContentType':
        """
        Returns the content of a ``Document`` object.

        :param doc: A docarray ``Document`` object.
        :return: Content of the input document.
        """
        return doc.content


class TextPreprocess(DefaultPreprocess):
    """
    Default built-in preprocess class to unpack the text from a ``Document`` object.
    """

    pass


class VisionPreprocess(BasePreprocess):
    """
    Built-in preprocess class for ``Document`` objects holding image data. It
    transforms an image given by a URI or a tensor into a tensor which can be passed
    into an image embedding model.
    If the image is passed as an int8 tensor or a URI, normalization is applied.
    If the image is passed as a float32 tensor, normalization is omitted.

    :param height: The target height of the image.
    :param width: The target width of the image.
    :param channel_axis: The default channel axis of the image. After preprocess will be
      set to 0 (C * H * W).
    :param augmentation: If `True`, apply random augmentation for the image content.
    """

    def __init__(
        self,
        height: int = 224,
        width: int = 224,
        channel_axis: int = -1,
        augmentation: bool = False,
    ):
        self._height = height
        self._width = width
        self._channel_axis = channel_axis
        self._augmentation = augmentation

    def __call__(self, doc: Document) -> np.ndarray:
        """
        Unpacks and preprocesses the content of a ``Document`` object with image
        content.

        :param doc: A docarray ``Document`` object.
        :return: Preprocessed tensor content of the input document.
        """
        current_channel_axis = self._channel_axis
        doc = Document(doc, copy=True)
        if doc.tensor is None:
            if doc.uri:
                doc.load_uri_to_image_tensor(
                    width=self._width,
                    height=self._height,
                    channel_axis=self._channel_axis,
                )
            else:
                raise AttributeError(
                    f'Document `tensor` is None, loading it from url: {doc.uri} failed.'
                )
        doc.set_image_tensor_shape(
            shape=(self._height, self._width), channel_axis=self._channel_axis
        )
        # Normalize image as np.float32.
        if doc.tensor.dtype in [np.int8, np.uint8]:
            doc.tensor = doc.tensor.astype(np.uint8)
            doc.set_image_tensor_normalization(channel_axis=self._channel_axis)
        elif doc.tensor.dtype == np.float64:
            doc.tensor = np.float32(doc.tensor)
        if self._augmentation:
            import albumentations as A

            if self._channel_axis not in _CHANNEL_LAST:
                # if image is not channel_last, move C to last axis.
                # This is required for albumentations.
                doc.set_image_tensor_channel_axis(self._channel_axis, _CHANNEL_LAST[0])
                current_channel_axis = _CHANNEL_LAST[0]

            transform = A.Compose(
                [
                    A.HorizontalFlip(p=0.5),
                    A.ColorJitter(p=1),
                    A.RandomResizedCrop(width=self._width, height=self._height, p=1),
                    A.GaussianBlur(p=1),
                    A.GridDropout(
                        ratio=0.2, p=0.5
                    ),  # random erase 0.2 percent of image with 0.5 probability
                ]
            )
            doc.tensor = transform(image=doc.tensor)['image']

        # Set image channel axis to pytorch default channel 0.
        doc.set_image_tensor_channel_axis(current_channel_axis, 0)

        return doc.tensor



import abc
from typing import TYPE_CHECKING, Any, List

import numpy as np

if TYPE_CHECKING:
    from docarray.typing import DocumentContentType


class BaseCollate(abc.ABC):
    @abc.abstractmethod
    def __call__(self, inputs: List['DocumentContentType']) -> Any:
        ...


class DefaultCollate(BaseCollate):
    """
    Default built-in collate class to create batch from a list of inputs for an
    embedding model.
    """

    def __call__(self, inputs: List['DocumentContentType']) -> Any:
        """
        Constructs a batch to pass into the embedding model from a list of contents.

        :param inputs: List of content objects to pass into the embedding model.
        :return: Any type of object that can be fed to an embedding model.
        """
        if isinstance(inputs[0], str):
            return inputs
        else:
            from torch.utils.data._utils.collate import default_collate

            return default_collate(inputs)


class TransformersCollate(BaseCollate):
    """
    Built-in collate class which applies a HuggingFace's transformer ``AutoTokenizer``
    on the given text inputs.

    :param name: The model id of a pretrained huggingface tokenizer or a path to a
        directory with the weights.
    :param padding: Set to True if padding should be applied during tokenization.
    :param kwargs: Keyword arguments to pass to the call of the ``AutoTokenizer``.
    """

    def __init__(self, name: str = 'bert-base-cased', padding: bool = True, **kwargs):
        from transformers import AutoTokenizer

        self._tokenizer = AutoTokenizer.from_pretrained(name)
        self._padding = padding
        self._kwargs = kwargs

    def __call__(self, inputs: List[str]):
        """
        Creates a ``BatchEncoding`` object from a list of input text values for the
        tokenizer.

        :param inputs: List of text values provied for the tokenization.
        :return: ``BatchEncoding`` objects to pass into a HuggingFace transformer
            embedding model.
        """
        return self._tokenizer(
            inputs, padding=self._padding, return_tensors='pt', **self._kwargs
        )


class VisionTransformersCollate(BaseCollate):
    """
    Built-in collate class which applies a HuggingFace's transformer
    ``AutoFeatureExtractor`` model on the given input image tensors.

    :param name: The model id of a pretrained HuggingFace model or a path to a
        directory with the model weights.
    :param kwargs: Keyword arguments to pass to the call of the feature extractor.
    """

    def __init__(self, name: str = 'openai/clip-vit-base-patch32', **kwargs):
        from transformers import AutoFeatureExtractor

        self._processor = AutoFeatureExtractor.from_pretrained(name)
        self._kwargs = kwargs

    def __call__(self, inputs: List[np.ndarray]):
        """
        Creates a ``BatchFeature`` object from a list of tensor values for the
        feature extractor.

        :param inputs: List tensors to pass into the feature extractor.
        :return: ``BatchFeature`` objects to pass into a vision transformer encoder
            model.
        """
        return self._processor(images=inputs, return_tensors='pt', **self._kwargs)


class OpenCLIPTextCollate(BaseCollate):
    """
    Built-in collate class which applies the tokenizer of OpenCLIP on the given text
    inputs.
    """

    def __init__(self):
        from open_clip import tokenize

        self._tokenize = tokenize

    def __call__(self, inputs: List[str]) -> 'torch.LongTensor':  # noqa: F821
        """
        Creates a tensor with a list of token ids for each text sequence passed to the
        collate function.
        """
        return self._tokenize(inputs)



from rich import box, print
from rich.table import Table


class DisplayMixin:
    def display(self, *args, **kwargs) -> None:
        """Display the model architecture from :py:attr:`.summary` in a table.

        :param args: args pass to :py:attr:`.summary`
        :param kwargs: kwargs pass to :py:attr:`.summary`
        """

        _summary = self.summary(*args, **kwargs)
        table = Table(box=box.SIMPLE)
        cols = ['name', 'output_shape_display', 'nb_params', 'trainable']
        for k in cols:
            table.add_column(k)
        for s in _summary:
            if s['trainable']:
                style = 'bright_green'
            else:
                style = 'cyan'
            if 'identity' in s['name']:
                style = 'bright_black'  # overwrite style if it is an identity layer.
            table.add_row(*map(str, (s[v] for v in cols)), style=style)
        print(
            table,
            '[green]Green[/green] layers are trainable layers, '
            '[cyan]Cyan[/cyan] layers are non-trainable layers or frozen layers.\n'
            '[bright_black]Gray[/bright_black] layers indicates this layer has been '
            'replaced by an Identity layer.\n'
            'Use to_embedding_model(...) to create embedding model.',
        )



import io
import logging
import os
import zipfile
from typing import Any, Dict, Optional, Tuple, Union
from zipfile import ZipFile

import hubble
import yaml

from .excepts import (
    ArtifactNotFound,
    CorruptedArtifact,
    CorruptedMetadata,
    NoSuchModel,
    SelectModelRequired,
)

MODEL_DUMP_FNAME = 'model.onnx'
MODEL_METADATA_FNAME = 'metadata.yml'

logger = logging.getLogger(__name__)


def _load_artifact_from_dir(
    artifact: str,
) -> Dict[str, Tuple[io.BytesIO, Union[str, io.BytesIO]]]:
    """
    Load artifact buffers from local dir.
    """

    _models2buffers: Dict[str, Tuple[io.BytesIO, Union[str, io.BytesIO]]] = {}

    for _root, dirs, files in os.walk(artifact):
        if MODEL_DUMP_FNAME in files and MODEL_METADATA_FNAME in files:
            _metadata_stream = io.BytesIO()
            with open(os.path.join(_root, MODEL_METADATA_FNAME), 'rb') as f:
                _metadata_stream.write(f.read())

            _models2buffers[_root] = (
                _metadata_stream,
                os.path.join(_root, MODEL_DUMP_FNAME),
            )

    return _models2buffers


def _read_local_zipfile(fname: str) -> io.BytesIO:
    """
    Read local zipfile to a buffer.
    """
    zbuffer = io.BytesIO()
    with open(fname, 'rb') as f:
        zbuffer.write(f.read())

    return zbuffer


def _pull_remote_artifact(artifact_id: str, token: Optional[str] = None) -> io.BytesIO:
    """
    Pull artifact from Hubble.
    """
    zbuffer = io.BytesIO()
    client = hubble.Client(jsonify=False, token=token)
    try:
        _ = client.download_artifact(id=artifact_id, f=zbuffer)
    except Exception as exc:
        raise ArtifactNotFound(f'Artifact {artifact_id} not found, details: {str(exc)}')

    return zbuffer


def _load_artifact_from_zipfile(
    artifact: str, zbuffer: io.BytesIO
) -> Dict[str, Tuple[io.BytesIO, Union[str, io.BytesIO]]]:
    """
    Load artifact buffers from zipfile.
    """

    _models2buffers: Dict[str, Tuple[io.BytesIO, Union[str, io.BytesIO]]] = {}

    zbuffer.seek(0)
    zf: ZipFile
    try:
        zf = ZipFile(zbuffer, mode='r', allowZip64=True)
    except zipfile.BadZipfile:
        raise CorruptedArtifact(f'Not a zip file: {artifact}')

    paths = zf.namelist()
    _intermediate_models2buffers = {}
    for path in paths:
        *header, tail = path.split('/')
        if tail == MODEL_METADATA_FNAME:
            prefix = '/'.join(header)
            _intermediate_models2buffers[prefix] = io.BytesIO(zf.read(path))

    for path in paths:
        *header, tail = path.split('/')
        prefix = '/'.join(header)
        if tail == MODEL_DUMP_FNAME and prefix in _intermediate_models2buffers:
            _metadata_stream = _intermediate_models2buffers[prefix]
            _models2buffers[prefix] = (
                _metadata_stream,
                io.BytesIO(zf.read(path)),
            )

    return _models2buffers


def load_artifact(
    artifact: str, select_model: Optional[str] = None, token: Optional[str] = None
) -> Tuple[str, Dict[str, Any], Union[str, io.BytesIO]]:
    """
    Load the artifact either from a directory, a zip file or directly from
    Hubble.
    """

    _models2buffers: Dict[str, Tuple[io.BytesIO, Union[str, io.BytesIO]]]

    if os.path.isdir(artifact):
        logger.info(f'Loading from local dir: \'{artifact}\'')
        _models2buffers = _load_artifact_from_dir(artifact)
    else:
        if os.path.isfile(artifact):
            logger.info(f'Loading from local file: \'{artifact}\'')
            zbuffer = _read_local_zipfile(artifact)
        else:
            logger.info(f'Pulling from Hubble: {artifact}')
            zbuffer = _pull_remote_artifact(artifact, token)

        _models2buffers = _load_artifact_from_zipfile(artifact, zbuffer)

    if len(_models2buffers) == 0:
        raise CorruptedArtifact(
            'Could not locate any directory with the '
            f'\'{MODEL_METADATA_FNAME}\' and \'{MODEL_DUMP_FNAME}\' files'
        )

    _model_metadata: Dict[str, Dict[str, Any]] = {}
    _model_dumps: Dict[str, Union[str, io.BytesIO]] = {}

    for prefix, (_meta_stream, _model_dump) in _models2buffers.items():

        _meta_stream.seek(0)
        try:
            metadata = yaml.safe_load(_meta_stream)
        except Exception as exc:
            raise CorruptedMetadata(
                f'Metadata found in {prefix} are corrupted, details: {str(exc)}'
            )

        _meta_stream.close()

        name = metadata.get('name')
        if name is None:
            raise CorruptedMetadata(f'Field \'name\' is missing from {prefix}')

        _model_metadata[name] = metadata
        _model_dumps[name] = _model_dump

    logger.info(f'Found models: {list(_model_metadata.keys())}')

    if select_model:
        if select_model not in _model_metadata:
            raise NoSuchModel(f'No model named \'{select_model}\' in artifact')
        model_name = select_model
    else:
        if len(_model_metadata) > 1:
            raise SelectModelRequired(
                'Found more than 1 models in artifact, please select model '
                'using `select_model`'
            )
        model_name = list(_model_metadata.keys())[0]

    model_metadata = _model_metadata[model_name]
    onnx_dump = _model_dumps[model_name]

    return model_name, model_metadata, onnx_dump



from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union

import torch
from docarray import Document
from torch.utils.data._utils.collate import default_collate

from .collate import BaseCollate, DefaultCollate
from .preprocess import BasePreprocess, DefaultPreprocess

if TYPE_CHECKING:
    from docarray.typing import DocumentContentType

    #: The type of preprocessing function
    PreprocessFnType = Union[BasePreprocess, Callable[[Document], DocumentContentType]]

    #: The type of collate function
    CollateFnType = Union[
        BaseCollate,
        Callable[
            [List[DocumentContentType]],
            Any,
        ],
    ]


__default_modality_key__ = 'input'


class PreprocessCollateWrapper:
    """
    This helper class wraps the preprocess and collate functions, that are provided
    by the user and are defined on single modalities, and exposes new preprocess-all
    and collate-all functions that can work on either single-modal or multi-modal
    cases. The class also handles the resolution of the input preprocess and collate
    functions.
    """

    def __init__(
        self,
        preprocess_fn: Union[
            Optional['PreprocessFnType'], Dict[str, Optional['PreprocessFnType']]
        ],
        collate_fn: Union[
            Optional['CollateFnType'], Dict[str, Optional['CollateFnType']]
        ],
        multi_modal: bool = False,
    ) -> None:
        """Constructor."""
        self._preprocess_fns: Dict[str, Optional['CollateFnType']]
        self._collate_fns: Dict[str, Optional['CollateFnType']]

        self._preprocess_fns, self._collate_fns = self.resolve(
            preprocess_fn, collate_fn
        )

        # This will be a list of unique elements since it is derived from a dict
        # Also the order will be preserved
        self._modalities = list(self._preprocess_fns.keys())

        if not multi_modal and len(self._modalities) > 1:
            raise ValueError(
                'You need to set multi_modal to True in order to process '
                f'{len(self._modalities)} modalities'
            )

        self._preprocess: Callable = (
            self._multi_modal_preprocess
            if multi_modal
            else self._single_modal_preprocess
        )

    def _single_modal_preprocess(
        self, doc: Document
    ) -> Dict[str, 'DocumentContentType']:
        """
        Document preprocess method for single modality cases.
        """
        return {  # this will always have a length of 1
            modality: func(doc) for modality, func in self._preprocess_fns.items()
        }

    def _multi_modal_preprocess(
        self, doc: Document
    ) -> Dict[str, 'DocumentContentType']:
        """
        Document preprocess method for multiple modality cases.
        """
        if len(doc.chunks) < len(self._modalities):
            raise ValueError(
                f'Found doc: {doc.id} with {len(doc.chunks)} chunks, expected '
                f'modalities are {len(self._modalities)}'
            )

        chunks_modalities = [chunk.modality for chunk in doc.chunks]
        if set(self._modalities).issubset(chunks_modalities):
            # resolve which chunk goes to which preprocess function using the
            # .modality attribute
            return {
                modality: func(
                    next(filter(lambda x: x.modality == modality, doc.chunks))
                )
                for modality, func in self._preprocess_fns.items()
            }

        # resolve which chunk goes to which preprocess function using the index,
        # i.e. first chunk goes to the first modality defined etc
        return {
            modality: func(doc.chunks[i])
            for i, (modality, func) in enumerate(self._preprocess_fns.items())
        }

    @property
    def modalities(self) -> List[str]:
        """
        Get the input modalities.
        """
        return self._modalities

    def preprocess(self, doc: Document) -> Dict[str, 'DocumentContentType']:
        """
        Exposed preprocess method. Wraps per-modality preprocess functions to a single
        preprocess function.
        """
        return self._preprocess(doc)

    def collate_contents(
        self, batch: List[Dict[str, 'DocumentContentType']]
    ) -> Dict[str, torch.Tensor]:
        """
        Exposed collate method for contents only.
        """
        return {
            modality: collate_fn([content[modality] for content in batch])
            for modality, collate_fn in self._collate_fns.items()
        }

    @staticmethod
    def collate_labels(batch: List[int]) -> torch.Tensor:
        """
        Exposed collate method for labels only.
        """
        return default_collate(batch)

    def collate(
        self, batch: List[Tuple[Dict[str, 'DocumentContentType'], int]]
    ) -> Tuple[Dict[str, torch.Tensor], torch.Tensor]:
        """
        Exposed collate method. Wraps per-modality collate functions to a single
        collate function.
        """
        contents, labels = zip(*batch)
        return self.collate_contents(contents), self.collate_labels(labels)

    @staticmethod
    def resolve(
        preprocess_fn: Union[
            Optional['PreprocessFnType'], Dict[str, Optional['PreprocessFnType']]
        ],
        collate_fn: Union[
            Optional['CollateFnType'], Dict[str, Optional['CollateFnType']]
        ],
    ) -> Tuple[Dict[str, 'PreprocessFnType'], Dict[str, 'CollateFnType']]:
        """
        Resolves the user provided preprocess and collate functions. The goal of this
        function is to have the resolution and the checking of the user provided
        preprocess and collate functions as well as the falling back to defaults, in one
        place. Given the user provided preprocess and collate functions, which can
        be functions or dicts, this method tries to resolve missing keys, incosistent
        keys as well as falling back to defaults when necessary.
        """
        preprocess_fns: Dict[str, 'PreprocessFnType']
        collate_fns: Dict[str, 'CollateFnType']

        if preprocess_fn is None:
            preprocess_fns = {__default_modality_key__: DefaultPreprocess()}
        elif isinstance(preprocess_fn, dict):
            if len(preprocess_fn) == 0:
                raise ValueError('Got an empty dict of preprocess functions')
            else:
                preprocess_fns = {
                    modality: func or DefaultPreprocess()
                    for modality, func in preprocess_fn.items()
                }
        else:
            preprocess_fns = {__default_modality_key__: preprocess_fn}

        if collate_fn is None:
            collate_fns = {__default_modality_key__: DefaultCollate()}
        elif isinstance(collate_fn, dict):
            if len(collate_fn) == 0:
                raise ValueError('Got an empty dict of collate functions')
            else:
                collate_fns = {
                    modality: func or DefaultCollate()
                    for modality, func in collate_fn.items()
                }
        else:
            collate_fns = {__default_modality_key__: collate_fn}

        preprocess_modalities = set(preprocess_fns.keys())
        collate_modalities = set(collate_fns.keys())

        if preprocess_modalities != collate_modalities:
            raise ValueError(
                f'Got different modalities for preprocess and collate functions, '
                f'{preprocess_modalities} != {collate_modalities}'
            )

        return preprocess_fns, collate_fns



import torch.nn as nn


class ProjectionHead(nn.Module):
    """Projection head used internally for self-supervised training.
    It is (by default) a simple 3-layer MLP to be attached on top of embedding model
    only for training purpose. After training, it should be cut-out from the embedding
    model.
    """

    EPSILON = 1e-5

    def __init__(self, in_features: int, output_dim: int = 128, num_layers: int = 2):
        super().__init__()
        self.head_layers = nn.ModuleList()
        for _ in range(num_layers - 1):
            self.head_layers.append(
                nn.Linear(in_features=in_features, out_features=in_features, bias=False)
            )
            self.head_layers.append(
                nn.BatchNorm1d(num_features=in_features, eps=self.EPSILON)
            )
            self.head_layers.append(nn.ReLU())
        self.head_layers.append(
            nn.Linear(in_features=in_features, out_features=output_dim, bias=False)
        )
        self.head_layers.append(
            nn.BatchNorm1d(num_features=output_dim, eps=self.EPSILON)
        )

    def forward(self, x):
        for layer in self.head_layers:
            x = layer(x)
        return x



import os

from setuptools import find_packages, setup

# package version
__version__ = '0.0.0'
try:
    libinfo_py = os.path.join('__init__.py')
    libinfo_content = open(libinfo_py, 'r', encoding='utf8').readlines()
    version_line = [
        line.strip() for line in libinfo_content if line.startswith('__version__')
    ][0]
    exec(version_line)  # gives __version__
except FileNotFoundError:
    pass

if __name__ == '__main__':
    setup(
        name='finetuner-commons',
        packages=find_packages(include=['commons*']),
        version='0.0.1b7',
        include_package_data=True,
        description='The finetuner-commons package provides common functionality between core and client.',
        author='Jina AI',
        author_email='team-finetuner@jina.ai',
        url='https://github.com/jina-ai/finetuner.fit/',
        license='Proprietary',
        download_url='https://github.com/jina-ai/finetuner.fit/tags',
        long_description_content_type='text/markdown',
        zip_safe=False,
        setup_requires=['setuptools>=18.0', 'wheel'],
        install_requires=[
            'docarray[common]>=0.13.25',
            'jina-hubble-sdk>=0.15.2',
            'onnxruntime==1.11.1',
            'rich>=12.4.4',
            'torchvision==0.13.0',
            'torch==1.12.0',
            'transformers==4.20.1',
        ],
        python_requires='>=3.7.0',
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'Intended Audience :: Education',
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 3.8',
            'Environment :: Console',
            'Operating System :: OS Independent',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
        ],
        project_urls={
            'Source': 'https://github.com/jina-ai/finetuner.fit/',
            'Tracker': 'https://github.com/jina-ai/finetuner.fit/issues',
        },
        keywords=(
            'jina neural-search neural-network deep-learning pretraining '
            'fine-tuning pretrained-models triplet-loss metric-learning '
            'siamese-network few-shot-learning'
        ),
    )



from typing import List, Optional, Tuple, Union

from torch import nn

from .projection import ProjectionHead  # noqa: F401
from .tailor import Tailor  # noqa: F401


def to_embedding_model(
    model: nn.Module,
    layer_name: Optional[str] = None,
    input_shape: Optional[Tuple[int, ...]] = None,
    input_dtype: str = 'float32',
    freeze: Union[bool, List[str]] = False,
    projection_head: Optional[nn.Module] = None,
) -> nn.Module:
    """Convert a general model from :py:attr:`.model` to an embedding model.

    :param model: The DNN model to be converted.
    :param layer_name: The name of the layer that is used for output embeddings. All
        layers *after* that layer will be removed. When set to ``None``, then the last
        layer listed in :py:attr:`.embedding_layers` will be used. To see all available
        names you can check ``name`` field of :py:attr:`.embedding_layers`.
    :param input_shape: The input shape of the DNN model.
    :param input_dtype: The input data type of the DNN model.
    :param freeze: If set as True, will freeze all layers before
        :py:`attr`:`layer_name`. If set as list of str, will freeze layers by names.
    :param projection_head: Attach a module at the end of model, this module should be
        always trainable.
    """
    return Tailor(model, input_shape, input_dtype).to_embedding_model(
        layer_name=layer_name,
        projection_head=projection_head,
        freeze=freeze,
    )


def display(
    model: nn.Module,
    input_shape: Optional[Tuple[int, ...]] = None,
    input_dtype: str = 'float32',
) -> None:
    """Display the model architecture from :py:attr:`.summary` in a table.

    :param model: The DNN model to display.
    :param input_shape: The input shape of the DNN model.
    :param input_dtype: The input data type of the DNN model.
    """
    Tailor(model, input_shape, input_dtype).display()
